package de.renew.net;

import de.renew.database.Transaction;
import de.renew.database.TransactionSource;

import de.renew.engine.common.SimulatorEventLogger;
import de.renew.engine.common.StepIdentifier;
import de.renew.engine.events.NetInstantiation;

import de.renew.unify.Impossible;

import de.renew.util.DelayedFieldOwner;
import de.renew.util.RenewObjectInputStream;
import de.renew.util.RenewObjectOutputStream;

import java.io.IOException;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;


public class NetInstanceImpl implements DelayedFieldOwner, NetInstance {
    public static org.apache.log4j.Logger logger = org.apache.log4j.Logger
                                                   .getLogger(NetInstanceImpl.class);
    static final long serialVersionUID = 6136525457931796546L;
    public static boolean useGlobalIDRegistry = false;
    private Hashtable instanceLookup;
    private Net net;
    private String netID;

    /**
     * The IDRegistry that is responsible for this net instance.
     * The registry is kept in a field, because the net instance
     * has to deregister its tokens at exactly this net instance
     * during finialization. The current registry might have
     * changed by that time.
     * <p>
     * This field is not really transient, but as we want
     * to cut down the serialization recursion depth, it
     * is serialized manually.</p>
     */
    private transient IDRegistry registry;

    protected NetInstanceImpl() {
    }

    protected NetInstanceImpl(Net net) throws Impossible {
        this(net, true);
    }

    protected NetInstanceImpl(Net net, boolean wantInitialTokens)
            throws Impossible {
        initNet(net, wantInitialTokens);
    }

    protected void initNet(Net net, boolean wantInitialTokens)
            throws Impossible {
        if (net == null) {
            throw new Impossible();
        }
        this.net = net;

        netID = IDSource.createID();

        if (useGlobalIDRegistry) {
            registry = IDRegistry.getInstance();
        } else {
            registry = new IDRegistry();
        }

        instanceLookup = new Hashtable();

        Iterator iterator = net.places().iterator();
        while (iterator.hasNext()) {
            Place place = (Place) iterator.next();
            instanceLookup.put(place,
                               place.makeInstance(this, wantInitialTokens));
        }

        iterator = net.transitions.iterator();
        while (iterator.hasNext()) {
            Transition transition = (Transition) iterator.next();
            instanceLookup.put(transition,
                               new TransitionInstance(this, transition));
        }
    }

    public String toString() {
        return net.toString() + "[" + netID + "]";
    }

    /**
     * Get the IDRegistry that is responsible for this net instance.
     */
    public IDRegistry getRegistry() {
        return registry;
    }

    /**
     * Query the ID of the net instance. By default the ID is a
     * simple number, but this is not guaranteed.
     *
     * @return the current ID string
     */
    public String getID() {
        return netID;
    }

    /**
     * Set the ID of the net instance. This should only be done
     * during the setup of the net instance, typeically after restoring
     * the net from a saved state.
     *
     * @param id the new ID string
     */
    public void setID(String id) {
        netID = id;
    }

    /**
     * The creation of the net instance is confirmed for the
     * purposes of the database storage. This confirmation can still
     * be rolled back. No actions except database accesses are performed
     * here.
     *
     * The access has to be done quite early, because
     * the net must be known to the database transaction.
     * Also, the initial tokens may have to be inserted early.
     */
    public void earlyConfirmation() {
        // First, the net is made public to the database.
        Transaction transaction = TransactionSource.get();
        try {
            transaction.createNet(this);
        } catch (Exception e) {
            logger.error(e.getMessage(), e);
        }

        // Afterwards, the places are notified, too.
        Iterator iterator = net.places().iterator();
        while (iterator.hasNext()) {
            Place place = (Place) iterator.next();
            getInstance(place).earlyConfirmation();
        }
    }

    /**
     * Trace the initial tokens of every place, if they are
     * inserted early on. Also, make sure that this net instance
     * is accessible via the net list, if this is desired.
     *
     */
    public void earlyConfirmationTrace(StepIdentifier stepIdentifier) {
        Iterator iterator = net.places().iterator();
        while (iterator.hasNext()) {
            Place place = (Place) iterator.next();
            getInstance(place).earlyConfirmationTrace(stepIdentifier);
        }

        NetInstanceList.add(this);
    }

    /**
     * Upon creation, no transition will be checked for enabledness
     * by a searcher, because it is not contained in the
     * search queue of possibly activated transitions.
     * By calling this method, all transitions are inserted into the search
     * queue. Then all tokens of the initial marking that might
     * have been held back are inserted into the places.
     *
     * Additionally, if the net's automatic registration flag is
     * set, the net will be registered at the net list.
     *
     * @see de.renew.net.Net#setAutomaticRegistration(boolean)
     * @see de.renew.net.NetInstanceList
     *
     */
    public void lateConfirmation(StepIdentifier stepIdentifier) {
        // log instance creation on net level
        SimulatorEventLogger.log(stepIdentifier, new NetInstantiation(this),
                                 this);


        // The places may now fully acquire their correct initial marking,
        // even if the initial marking is inserted late.
        Iterator iterator = net.places().iterator();
        while (iterator.hasNext()) {
            Place place = (Place) iterator.next();
            getInstance(place).lateConfirmation(stepIdentifier);
        }


        // Let's inform all transitions of the confirmation.
        // The transitions can insert themselves into the search queue.
        iterator = net.transitions().iterator();
        while (iterator.hasNext()) {
            Transition transition = (Transition) iterator.next();
            getInstance(transition).createConfirmation();
        }
    }

    /**
     * Call this method if the creation of a net instance can be confirmed
     * in one single step.
     *
     */
    public void createConfirmation(StepIdentifier stepIdentifier) {
        earlyConfirmation();
        earlyConfirmationTrace(stepIdentifier);
        lateConfirmation(stepIdentifier);
    }

    public Object getInstance(Object netObject) {
        return instanceLookup.get(netObject);
    }

    public PlaceInstance getInstance(Place place) {
        return (PlaceInstance) instanceLookup.get(place);
    }

    public TransitionInstance getInstance(Transition transition) {
        return (TransitionInstance) instanceLookup.get(transition);
    }

    public Net getNet() {
        return net;
    }

    public Collection getUplinkProviders(String channel) {
        List result = new ArrayList();

        Iterator iterator = getNet().transitions().iterator();
        while (iterator.hasNext()) {
            Transition transition = (Transition) iterator.next();
            TransitionInstance instance = getInstance(transition);
            if (instance.listensToChannel(channel)) {
                result.add(instance);
            }
        }
        return result;
    }

    /**
     * Serialization method, behaves like default writeObject
     * method except using the domain trace feature, if the
     * output is a {@link RenewObjectOutputStream}.
     * @see de.renew.util.RenewObjectOutputStream
     **/
    private void writeObject(java.io.ObjectOutputStream out)
            throws IOException {
        RenewObjectOutputStream rOut = null;
        if (out instanceof RenewObjectOutputStream) {
            rOut = (RenewObjectOutputStream) out;
        }
        if (rOut != null) {
            rOut.beginDomain(this);
        }
        out.defaultWriteObject();
        if (rOut != null) {
            rOut.delayedWriteObject(registry, this);
            rOut.endDomain(this);
        } else {
            out.writeObject(registry);
        }
    }

    /**
     * Deserialization method, behaves like default readObject
     * method except restoring additional transient fields.
     * The method restores the not-really-transient field
     * <code>registry</code>, <b>if</b> the stream used is
     * <b>not</b> a {@link RenewObjectInputStream}.
     * The method also repeats this instance's registration at the
     * {@link NetInstanceList}.
     **/
    private void readObject(java.io.ObjectInputStream in)
            throws IOException, ClassNotFoundException {
        in.defaultReadObject();

        if (in instanceof RenewObjectInputStream) {
            // Do nothing, the fields will be
            // reassigned by the stream soon.
        } else {
            registry = (IDRegistry) in.readObject();
        }

        // This is normally done at the net instance's creation
        // confirmation, but the confirmation of this instance's creation
        // has taken place long ago and far away.
        // So we have to repeat it here, in our current universe.
        NetInstanceList.add(this);
    }

    /**
     * Deserialization method used by {@link RenewObjectInputStream}.
     * Reassigns a value to the not-really-transient field
     * <code>registry</code>.
     *
     * @exception java.io.NotSerializableException
     *    if the given value was not an IDRegistry object.
     */
    public void reassignField(Object value) throws IOException {
        if (value instanceof IDRegistry) {
            registry = (IDRegistry) value;
        } else {
            throw new java.io.NotSerializableException("Value of unexpected type given to NetInstanceImpl.reassignField():"
                                                       + value.getClass()
                                                              .getName() + ".");
        }
    }
    // ### Problem: NetInstanceReference only used for database
    // serialization, not during normal simulation save.
    //    /**
    //     * Serialization method to replace the object
    //     * before it is serialized.
    //     * In this case, the NetInstanceImpl is replaced
    //     * by a NetInstanceImplReference.
    //     * @return A NetInstanceImplReference pointing to
    //     * the original NetInstanceImpl by its NetID.
    //     */
    //    private Object writeReplace()
    //    {
    //      return new NetInstanceReference(netID);
    //    }
}