package de.renew.unify;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;


// On the fly:
//   Check that no calculated unknown is bound.
//   Check that no calculated unknown is registered twice.
//   Check that no two calculated unknowns depend on each other.
// During the consistency check:
//   Check that all early variables have become bound.
//   Check that all late variables are either bound or calculated.
final public class CalculationChecker {
    private Set lateVariables;
    private Set earlyVariables;

    public CalculationChecker() {
        lateVariables = new HashSet();
        earlyVariables = new HashSet();
    }

    public void reset() {
        lateVariables.clear();
        earlyVariables.clear();
    }

    public void addEarlyVariable(final Variable var, StateRecorder recorder)
            throws Impossible {
        if (!earlyVariables.contains(var)) {
            if (recorder != null) {
                recorder.record(new StateRestorer() {
                        public void restore() {
                            earlyVariables.remove(var);
                        }
                    });
            }

            earlyVariables.add(var);
        }
    }

    public void addCalculated(Class targetType, Object target, Object source,
                              StateRecorder recorder) throws Impossible {
        Unify.unify(target, new Calculator(targetType, source, recorder),
                    recorder);
    }

    public void addLateVariable(final Variable var, StateRecorder recorder)
            throws Impossible {
        if (!lateVariables.contains(var)) {
            if (recorder != null) {
                recorder.record(new StateRestorer() {
                        public void restore() {
                            lateVariables.remove(var);
                        }
                    });
            }

            lateVariables.add(var);
        }
    }

    private boolean checkLateVariables() {
        // All late variables must be bound or calculated.
        Iterator iterator = lateVariables.iterator();
        while (iterator.hasNext()) {
            Variable variable = (Variable) iterator.next();
            if (!variable.isComplete()) {
                return false;
            }
        }
        return true;
    }

    private boolean checkEarlyVariables() {
        // All early variables must be bound.
        Iterator enumeration = earlyVariables.iterator();
        while (enumeration.hasNext()) {
            Variable variable = (Variable) enumeration.next();
            if (!variable.isBound()) {
                return false;
            }
        }
        return true;
    }

    public boolean isConsistent() {
        return checkLateVariables() && checkEarlyVariables();
    }
}